import sqlite3
from typing import Any

from dbtogo.datatypes import DBEngine, SQLColumn
from dbtogo.exceptions import DestructiveMigrationError
from dbtogo.migrations import Migration, MigrationEngine

sqlite_column = tuple[int, str, str, int, Any, int]


class SQLiteEngineError(Exception):
    pass


class SqliteEngine(DBEngine):
    def __init__(self, conn: sqlite3.Connection):
        self.conn = conn
        self.cursor = conn.cursor()

    def __del__(self) -> None:
        self.conn.close()

    def _represent_bytes(self, data: bytes) -> str:
        return f"X'{data.hex().upper()}'"

    def select(
        self, field: str, table: str, conditions: dict[str, Any] | None = None
    ) -> list[Any]:
        if conditions is None:
            query = f"SELECT {field} FROM {table}"
            self.cursor.execute(query)

        else:
            where_clause = " AND ".join(f"{key} = ?" for key in conditions.keys())
            query = f"SELECT {field} FROM {table} WHERE {where_clause}"
            self.cursor.execute(query, tuple(conditions.values()))

        return self.cursor.fetchall()

    def insert(self, table: str, obj_data: dict[str, Any]) -> int | None:
        cols = [col for col, val in obj_data.items() if val is not None]
        vals = [val for val in obj_data.values() if val is not None]

        col_str = ", ".join(cols)
        val_str = ", ".join(["?"] * len(vals))

        query = f"INSERT INTO {table} ({col_str}) VALUES({val_str})"

        self.cursor.execute(query, tuple(vals))
        self.conn.commit()
        return self.cursor.lastrowid

    def _transfer_type_from_standard(self, str_type: str) -> str:
        types = {
            "integer": "INTEGER",
            "date-time": "TIMESTAMP",
            "string": "TEXT",
            "number": "REAL",
            "boolean": "BOOLEAN",
            "bytes": "BLOB",
        }

        return types[str_type]

    def _transfer_type_to_standard(self, str_type: str) -> str:
        types = {
            "INTEGER": "integer",
            "TIMESTAMP": "date-time",
            "TEXT": "string",
            "REAL": "number",
            "BOOLEAN": "boolean",
            "BLOB": "bytes",
        }

        return types[str_type]

    def _create_table(self, tablename: str, standard_cols: list[SQLColumn]) -> None:
        sqlite_cols = []
        for column in standard_cols:
            lite_col = (
                f"{column.name} {self._transfer_type_from_standard(column.datatype)}"
            )

            if column.nullable and not column.primary_key:
                lite_col += " NULLABLE"
            else:
                lite_col += " NOT NULL"

            if column.primary_key:
                lite_col += " PRIMARY KEY AUTOINCREMENT"

            if column.unique:
                lite_col += " UNIQUE"

            if column.default is not None:
                if column.datatype == "string":
                    lite_col += (
                        f" DEFAULT '{column.default.replace("'", '').replace('"', '')}'"
                    )
                elif column.datatype != "bytes":
                    lite_col += f" DEFAULT {column.default}"
                else:
                    lite_col += f" DEFAULT {self._represent_bytes(column.default)}"

            sqlite_cols.append(lite_col)

        query = f"CREATE TABLE IF NOT EXISTS {tablename} ({','.join(sqlite_cols)})"
        self.cursor.execute(query)

        self.conn.commit()

    def _drop_table(self, table: str) -> None:
        query = f"DROP TABLE IF EXISTS {table}"
        self.cursor.execute(query)
        self.conn.commit()

    def _rename_table(self, old_table: str, new_table: str) -> None:
        query = f"ALTER TABLE {old_table} RENAME TO {new_table}"
        self.cursor.execute(query)
        self.conn.commit()

    def _parse_raw_column(self, column: str) -> SQLColumn:
        column_data: list[str] = column.split(" ")
        column_name = column_data.pop(0)
        column_type = self._transfer_type_to_standard(column_data.pop(0))

        nullable = False
        primary = False
        default = None
        unique = False

        while len(column_data) > 0:
            current = column_data.pop(0)
            match current:
                case "NOT":
                    column_data.pop(0)
                    nullable = False
                case "NULLABLE":
                    nullable = True
                case "DEFAULT":
                    default = column_data.pop(0)
                    if column_type == "string":
                        default = default.strip("'").strip('"')
                case "PRIMARY" | "KEY" | "AUTOINCREMENT":
                    primary = True
                case "UNIQUE":
                    unique = True
                case _:
                    raise SQLiteEngineError("Failed parsing current DB schema")

        return SQLColumn(column_name, column_type, nullable, default, primary, unique)

    def _get_SQLColumns(self, table: str) -> list[SQLColumn]:
        query = f"SELECT sql FROM sqlite_master WHERE type='table' AND name='{table}';"
        self.cursor.execute(query)
        current_schema = self.cursor.fetchone()[0]
        clean_schema: str = current_schema.split("(")[1].split(")")[0]
        raw_cols = clean_schema.split(",")

        standard_cols = []
        for raw_col in raw_cols:
            standard_cols.append(self._parse_raw_column(raw_col))

        return standard_cols

    def execute_migration(
        self,
        migration: Migration,
        force: bool = False,
        _current_cols: list[SQLColumn] | None = None,
    ) -> None:
        me = MigrationEngine()

        if len(migration.steps) < 1:
            return

        if not force and migration.is_destructive():
            raise DestructiveMigrationError()

        if _current_cols is None:
            _current_cols = self._get_SQLColumns(migration.table)

        new_cols = me.get_migrated_cols(_current_cols, migration)

        temp_table = f"_temp_migrate_{migration.table}"
        self._create_table(temp_table, new_cols)

        current = self.select("*", migration.table, None)
        val_string = ", ".join(["?"] * len(_current_cols))

        renamed = me.get_renamed_mapping(migration)
        not_dropped = [x.name for x in new_cols]
        col_str = ", ".join(
            [
                renamed.get(x.name, x.name)
                for x in _current_cols
                if renamed.get(x.name, x.name) in not_dropped
            ]
        )

        try:
            for row in current:
                query = f"INSERT INTO {temp_table} ({col_str}) VALUES({val_string})"
                self.cursor.execute(query, row)

            self.conn.commit()

        except Exception as e:
            self._drop_table(temp_table)
            raise e

        self._drop_table(migration.table)
        self._rename_table(temp_table, migration.table)

    def _migrate_from(self, table: str, new_columns: list[SQLColumn]) -> None:
        for col in new_columns:
            if col.datatype == "bytes" and col.default is not None:
                col.default = self._represent_bytes(col.default)

        current_columns = self._get_SQLColumns(table)

        migration = MigrationEngine().generate_migration(
            table, current_columns, new_columns
        )
        if len(migration.steps) > 0:
            self.execute_migration(migration)

    def migrate(self, table: str, columns: list[SQLColumn]) -> None:
        matched_tables = self.cursor.execute(
            f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table}';"
        ).fetchall()

        if len(matched_tables) == 0:
            return self._create_table(table, columns)

        else:
            return self._migrate_from(table, columns)

    def update(self, table: str, obj_data: dict[str, Any], primary_key: str) -> None:
        cols = [col for col, val in obj_data.items() if val is not None]
        vals = [val for val in obj_data.values() if val is not None]

        set_string = ""
        for col in cols:
            set_string += f"{col} = ?, "
        set_string = set_string[:-2]

        query = f"UPDATE {table} SET {set_string} WHERE {primary_key} = ?"
        vals.append(obj_data[primary_key])

        self.cursor.execute(query, tuple(vals))
        self.conn.commit()

    def delete(self, table: str, key: str, value: Any) -> None:
        query = f"DELETE FROM {table} WHERE {key} = ?"
        self.cursor.execute(query, (value,))
        self.conn.commit()



class NoBindError(Exception):
    def __init__(self) -> None:
        super().__init__("You need to bind this model to a db before first use")


class BindViolationError(Exception):
    def __init__(self) -> None:
        super().__init__("You cannot change the primary key of a bound object")


class UnboundDeleteError(Exception):
    def __init__(self) -> None:
        super().__init__("You cannot delete an object that is not in the db")


class AutoMigrationError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "dbtogo cant automatically migrate the schemas, consider manual migration"
        )


class DestructiveMigrationError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "This migration includes destructive actions,"
            "but wasnt executed with force=True"
        )


class InvalidMigrationError(Exception):
    def __init__(self) -> None:
        super().__init__("This migration is not valid")



import sqlite3
from collections.abc import Callable
from typing import Any, Self

from pydantic import BaseModel

from dbtogo.datatypes import DBEngine, SQLColumn, UnboundEngine
from dbtogo.exceptions import BindViolationError, NoBindError, UnboundDeleteError
from dbtogo.serialization import GeneralSQLSerializer
from dbtogo.sqlite import SqliteEngine

DEFAULT_PRIM_KEYS = ["id", "primary_key", "uuid"]


class DBEngineFactory:
    @staticmethod
    def create_sqlite3_engine(database: str = "") -> DBEngine:
        conn = sqlite3.connect(database)
        return SqliteEngine(conn)


def bound(func: Callable) -> Callable:
    def wrapper(cls: "DBModel", *args: tuple, **kwargs: dict[str, Any]) -> Any:
        if cls._db is None:
            raise NoBindError()

        return func(cls, *args, **kwargs)

    return wrapper


class DBModel(BaseModel):
    _db: DBEngine = UnboundEngine()
    _table: str = "table_not_set"
    _primary: str = "primary_not_set"

    @classmethod
    def bind(
        cls,
        db: DBEngine,
        primary_key: str | None = None,
        unique: list[str] = [],
        table: str | None = None,
    ) -> None:
        cls._db = db
        table = table if table is not None else cls.__name__

        columns = GeneralSQLSerializer().serialize_schema(
            cls.__name__, cls.model_json_schema(), primary_key, unique
        )

        used_names = [x.name for x in columns]
        if primary_key is None:
            for prim in DEFAULT_PRIM_KEYS:
                if prim in used_names:
                    continue
                primary_key = prim
                columns.append(SQLColumn(prim, "int", False, None, True, True))

        assert primary_key is not None

        cls._primary = primary_key
        cls._table = table
        db.migrate(table, columns)

    @classmethod
    def _deserialize_object(cls, object_data: tuple) -> Self:
        py_object = GeneralSQLSerializer().deserialize_object(cls, object_data)

        pk = getattr(py_object, py_object.__class__._primary)
        py_object._data_bind = pk

        return py_object

    @classmethod
    @bound
    def get(cls, **kwargs: dict[str, Any]) -> Self | None:
        data = cls._db.select("*", cls._table, kwargs)

        if len(data) < 1:
            return None

        return cls._deserialize_object(data[0])

    def _create(self) -> None:
        obj_data = GeneralSQLSerializer().serialize_object(self)

        insert_bind = self._db.insert(self.__class__._table, obj_data)
        bind_attr = getattr(self, self.__class__._primary)

        data_bind = bind_attr if bind_attr is not None else insert_bind
        self._data_bind = data_bind

    def _update(self) -> None:
        bind = self._data_bind

        if getattr(self, self.__class__._primary) != bind:
            raise BindViolationError()

        obj_data = GeneralSQLSerializer().serialize_object(self)
        self._db.update(self.__class__._table, obj_data, self.__class__._primary)

    @bound
    def save(self) -> None:
        if getattr(self, "_data_bind", None) is None:
            return self._create()
        return self._update()

    @bound
    def delete(self) -> None:
        if getattr(self, "_data_bind", None) is None:
            raise UnboundDeleteError()

        primary_key = self.__class__._primary
        primary_value = self._data_bind

        self._db.delete(self.__class__._table, primary_key, primary_value)
        self._data_bind = None

    @classmethod
    @bound
    def all(cls) -> list[Self]:
        data = cls._db.select("*", cls._table)
        return [cls._deserialize_object(x) for x in data]



from dbtogo.datatypes import (
    AddCol,
    AddConstraint,
    Migration,
    RemoveConstraint,
    RenameCol,
    SQLColumn,
    SQLConstraint,
    SQLType,
)
from dbtogo.dbmodel import DBEngineFactory, DBModel


class MigrationTestModelOld(DBModel):
    pk: int | None = None
    unq_string: str = "asdf"
    nullable_int: int | None = None
    modify_me: int

    @classmethod
    def bind(cls, engine):
        super().bind(
            engine,
            primary_key="pk",
            unique=["unq_string", "modify_me"],
            table="migration_test",
        )


class MigrationTestModelNew(DBModel):
    pk: int | None = None
    unq_string: str | None = "q"
    the_same_int: int | None = None
    modify_me: int = 7
    new_col: str | None = "default"

    @classmethod
    def bind(cls, engine):
        super().bind(
            engine,
            primary_key="pk",
            unique=["unq_string"],
            table="migration_test",
        )


def test_automatic_migration():
    engine = DBEngineFactory.create_sqlite3_engine("test.db")
    MigrationTestModelOld.bind(engine)

    a = MigrationTestModelOld(unq_string="hello", modify_me=8)
    b = MigrationTestModelOld(modify_me=5, nullable_int=1)

    a.save()
    b.save()

    MigrationTestModelNew.bind(engine)
    engine._drop_table("migration_test")


def test_manual_migration():
    engine = DBEngineFactory.create_sqlite3_engine("test.db")
    MigrationTestModelOld.bind(engine)

    a = MigrationTestModelOld(unq_string="hello", modify_me=8)
    b = MigrationTestModelOld(modify_me=5, nullable_int=1)

    a.save()
    b.save()

    valid_migration_steps = [
        RenameCol("nullable_int", "the_same_int"),
        RenameCol("modify_me", "i_am_modified"),
        AddConstraint("modify_me", SQLConstraint.nullable.value),
        RemoveConstraint("modify_me", SQLConstraint.unique.value),
        AddCol(SQLColumn("new_col", SQLType.string.value, True, "default")),
    ]

    valid_migration = Migration("migration_test", valid_migration_steps)

    engine.execute_migration(valid_migration, False)

    a.delete()
    b.delete()
    engine._drop_table("migration_test")



from dbtogo.dbmodel import DBEngineFactory, DBModel


class TestModel(DBModel):
    pk: int | None = None
    unq_string: str
    nullable_int: int | None = None
    bytes_test: bytes | None = None
    children: list["TestModel"] = []

    @classmethod
    def bind(cls, engine):
        super().bind(
            engine,
            primary_key="pk",
            unique=["unq_string"],
            table="test_edges",
        )


def test_edges():
    engine = DBEngineFactory.create_sqlite3_engine("test.db")
    TestModel.bind(engine)

    obj1 = TestModel(
        unq_string="OBJ1",
    )

    obj2 = TestModel(unq_string="OBJ2", nullable_int=5, bytes_test=b"deadbeef")

    obj1.children.append(obj1)
    obj1.save()

    obj2.save()

    obj1 = TestModel.get(unq_string="OBJ1")
    obj2 = TestModel.get(unq_string="OBJ2")

    assert len(obj1.children) != 0
    assert len(obj2.children) == 0

    obj1.delete()
    obj2.delete()

    assert len(TestModel.all()) == 0



from copy import deepcopy

from dbtogo.datatypes import (
    AddCol,
    AddConstraint,
    ChangeDefault,
    DropCol,
    Migration,
    MigrationStep,
    RemoveConstraint,
    RenameCol,
    RetypeCol,
    SQLConstraint,
)
from dbtogo.exceptions import InvalidMigrationError
from dbtogo.serialization import SQLColumn


class MigrationEngine:
    def _get_contraint_diff(
        self, old: SQLColumn, new: SQLColumn, constraint: str
    ) -> MigrationStep | None:
        if getattr(old, constraint) == getattr(new, constraint):
            return None

        if getattr(old, constraint):
            return RemoveConstraint(new.name, SQLConstraint.nullable.value)

        return AddConstraint(new.name, SQLConstraint.nullable.value)

    def get_col_diff(
        self, old_col: SQLColumn, new_col: SQLColumn
    ) -> list[MigrationStep]:
        steps: list[MigrationStep] = []

        if old_col.datatype != new_col.datatype:
            steps.append(RetypeCol(new_col.name, old_col.datatype, new_col.datatype))

        if old_col.default != new_col.default:
            steps.append(ChangeDefault(new_col.name, new_col.default))

        for constraint in ["nullable", "unique", "primary_key"]:
            step = self._get_contraint_diff(old_col, new_col, constraint)
            if step is not None:
                steps.append(step)

        return steps

    def generate_migration(
        self, table: str, original: list[SQLColumn], new: list[SQLColumn]
    ) -> Migration:
        steps = []

        current_names = [x.name for x in original]
        matched_cols = [x for x in new if x.name in current_names]

        for new_col in matched_cols:
            og_col = [x for x in original if x.name == new_col.name][0]
            steps += self.get_col_diff(og_col, new_col)

        matched_names = [x.name for x in matched_cols]

        added = [x for x in new if x.name not in matched_names]
        removed = [x for x in original if x.name not in matched_names]
        removed_sgn = [x.signature() for x in removed]

        for added_col in added:
            sgn = added_col.signature()
            if removed_sgn.count(sgn) == 1:
                rem_pos = removed_sgn.index(sgn)
                steps.append(RenameCol(removed[rem_pos].name, added_col.name))

                removed_sgn.pop(rem_pos)
                removed.pop(rem_pos)

            else:
                steps.append(AddCol(added_col))

        for to_be_dropped in removed:
            steps.append(DropCol(to_be_dropped.name))

        result = Migration(table, steps)
        result.sort()

        return result

    def get_renamed_mapping(self, migration: Migration) -> dict[str, str]:
        mapping = {}
        for step in migration.steps:
            if type(step) is RenameCol:
                mapping[step.old_name] = step.new_name
        return mapping

    def _execute_step(self, new_cols: dict[str, SQLColumn], step: MigrationStep) -> None:
        if type(step) is AddCol:
            new_cols[step.column.name] = step.column

        elif type(step) is DropCol:
            new_cols.pop(step.column_name)

        elif type(step) is RenameCol:
            col = new_cols.pop(step.old_name)
            col.name = step.new_name
            new_cols[col.name] = col

        elif type(step) is RetypeCol:
            new_cols[step.column_name].datatype = step.new_type

        elif type(step) is AddConstraint:
            col = new_cols[step.column_name]

            if step.constraint == SQLConstraint.nullable.value:
                col.nullable = True

            if step.constraint == SQLConstraint.unique.value:
                col.unique = True

            if step.constraint == SQLConstraint.primary.value:
                col.primary_key = True

        elif type(step) is RemoveConstraint:
            col = new_cols[step.column_name]

            if step.constraint == SQLConstraint.nullable.value:
                col.nullable = False

            if step.constraint == SQLConstraint.unique.value:
                col.unique = False

            if step.constraint == SQLConstraint.primary.value:
                col.primary_key = False

        elif type(step) is ChangeDefault:
            new_cols[step.column_name].default = step.new_default

    def get_migrated_cols(
        self, original: list[SQLColumn], migration: Migration
    ) -> list[SQLColumn]:
        new_cols = {}
        for col in original:
            new_cols[col.name] = deepcopy(col)

        migration.sort()

        for step in migration.steps:
            self._execute_step(new_cols, step)

        cols: list[SQLColumn] = list(new_cols.values())

        if len([x for x in cols if x.primary_key]) != 1:
            raise InvalidMigrationError()

        return cols



from dbtogo.dbmodel import DBEngineFactory, DBModel


class TestModel(DBModel):
    pk: int | None = None
    unq_string: str
    nullable_int: int | None = None
    bytes_test: bytes | None = None

    @classmethod
    def bind(cls, engine):
        super().bind(
            engine,
            primary_key="pk",
            unique=["unq_string"],
            table="test_basic",
        )


def test_crud():
    engine = DBEngineFactory.create_sqlite3_engine("test.db")
    TestModel.bind(engine)

    obj1 = TestModel(
        unq_string="OBJ1",
    )
    obj1.save()

    obj2 = TestModel(unq_string="OBJ2", nullable_int=5, bytes_test=b"deadbeef")
    obj2.save()

    obj1 = TestModel.get(unq_string="OBJ1")
    obj2 = TestModel.get(unq_string="OBJ2")

    assert obj2.bytes_test == b"deadbeef"

    assert obj1.unq_string == "OBJ1"
    assert obj2.unq_string == "OBJ2"

    assert obj1.pk is not None
    assert obj2.pk is not None

    assert obj1.nullable_int is None
    assert obj2.nullable_int == 5

    obj1.unq_string = "New_string"
    obj1.save()

    obj1 = TestModel.get(pk=obj1.pk)
    assert obj1.unq_string == "New_string"

    assert len(TestModel.all()) == 2

    obj1.delete()
    obj2.delete()

    assert len(TestModel.all()) == 0



from dbtogo.dbmodel import DBEngineFactory, DBModel


class Duck(DBModel):
    duck_id: int | None = None
    name: str
    color: str = "Brown"
    age: int | None = None
    shopping_list: list[str] = ["bread crumbs"]
    children: list["Duck"] = []

    @classmethod
    def bind(cls, engine):
        super().bind(engine, primary_key="duck_id", unique=["name"])

    def quack(self):
        print(f"Hi! I am {self.name} age {self.age} quack!")


def main():
    engine = DBEngineFactory.create_sqlite3_engine("test.db")

    Duck.bind(engine)

    mc_duck_junior = Duck(
        name="Junior",
        age=15,
        shopping_list=["rohlik", "gothaj"],
    )
    mc_duck_junior.save()

    mc_duck = Duck(name="McDuck", color="Yellow", age=45, children=[mc_duck_junior])
    mc_duck.save()

    mc_duck = Duck.get(name="McDuck")
    mc_duck.children[0].quack()

    mc_duck.name = "McDuckyDuck"
    mc_duck.save()

    mc_duck = Duck.get(duck_id=mc_duck.duck_id)
    print(mc_duck.name)

    print([x.name for x in Duck.all()])

    mc_duck.delete()
    mc_duck_junior.delete()


if __name__ == "__main__":
    main()



import abc
from enum import Enum
from typing import Any

from dbtogo.exceptions import NoBindError


class SQLType(Enum):
    integer = "integer"
    date_time = "date-time"
    string = "string"
    number = "number"
    boolean = "boolean"
    byte_data = "bytes"


class SQLConstraint(Enum):
    primary = "primary"
    nullable = "nullable"
    unique = "unique"


class SQLColumn:
    def __init__(
        self,
        name: str,
        datatype: str,
        nullable: bool,
        default: Any,
        primary: bool = False,
        unique: bool = False,
    ):
        self.name: str = name
        self.datatype: str = datatype
        self.nullable: bool = nullable
        self.default: Any = default
        self.primary_key: bool = primary
        self.unique: bool = unique

    def __str__(self) -> str:
        representation = f"{self.name}: "
        representation += "nullable " if self.nullable else ""
        representation += "unique " if self.unique else ""
        representation += "primary " if self.primary_key else ""
        representation += f"{self.datatype} ({self.default})"
        return representation

    def signature(self) -> str:
        str_signature = f"{self.datatype}{self.nullable}{self.default}"
        str_signature += f"{self.primary_key}{self.unique}"
        return str_signature


class MigrationStep:
    _destructive: bool = False


class AddCol(MigrationStep):
    def __init__(self, column: SQLColumn):
        self.column = column

    def __str__(self) -> str:
        return f"ADD {self.column.name}"


class DropCol(MigrationStep):
    def __init__(self, column_name: str):
        self.column_name = column_name
        self._destructive = True

    def __str__(self) -> str:
        return f"DROP {self.column_name}"


class RenameCol(MigrationStep):
    def __init__(self, old_name: str, new_name: str):
        self.old_name = old_name
        self.new_name = new_name

    def __str__(self) -> str:
        return f"RENAME {self.old_name} to {self.new_name}"


class RetypeCol(MigrationStep):
    def __init__(self, column_name: str, old_type: str, new_type: str):
        self.column_name = column_name
        self.old_type = old_type
        self.new_type = new_type
        self._destructive = True

    def __str__(self) -> str:
        return f"RETYPE {self.column_name} from {self.old_type} to {self.new_type}"


class AddConstraint(MigrationStep):
    def __init__(self, column_name: str, constraint: str):
        self.column_name = column_name
        self.constraint = constraint

        if constraint == SQLConstraint.primary.value:
            self._destructive = True

    def __str__(self) -> str:
        return f"ADD {self.constraint} to {self.column_name}"


class RemoveConstraint(MigrationStep):
    def __init__(self, column_name: str, constraint: str):
        self.column_name = column_name
        self.constraint = constraint

        if constraint == SQLConstraint.primary.value:
            self._destructive = True

    def __str__(self) -> str:
        return f"REMOVE {self.constraint} from {self.column_name}"


class ChangeDefault(MigrationStep):
    def __init__(self, column_name: str, new: Any):
        self.column_name = column_name
        self.new_default = new

    def __str__(self) -> str:
        return f"DEFAULT {self.column_name} to {self.new_default}"


class Migration:
    def __init__(self, table: str, steps: list[MigrationStep]):
        self.table = table
        self.steps = steps

    def is_destructive(self) -> bool:
        return len([x for x in self.steps if x._destructive]) > 0

    @staticmethod
    def _step_key_function(step: MigrationStep) -> int:
        if type(step) is AddCol:
            return 1

        elif type(step) is RetypeCol:
            return 3

        elif type(step) is AddConstraint:
            return 3

        elif type(step) is RemoveConstraint:
            return 3

        elif type(step) is DropCol:
            return 4

        elif type(step) is RenameCol:
            return 5

        return 0

    def sort(self) -> None:
        self.steps.sort(key=self._step_key_function)


class DBEngine(abc.ABC):
    @abc.abstractmethod
    def select(
        self, field: str, table: str, conditions: dict[str, Any] | None = None
    ) -> list[Any]:
        pass

    @abc.abstractmethod
    def insert(self, table: str, obj_data: dict[str, Any]) -> int | None:
        pass

    @abc.abstractmethod
    def migrate(self, table: str, columns: list[SQLColumn]) -> None:
        pass

    @abc.abstractmethod
    def update(self, table: str, obj_data: dict[str, Any], primary_key: str) -> None:
        pass

    @abc.abstractmethod
    def delete(self, table: str, key: str, value: Any) -> None:
        pass

    @abc.abstractmethod
    def execute_migration(self, migration: Migration, force: bool = False) -> None:
        pass


class UnboundEngine(DBEngine):
    def select(
        self, field: str, table: str, conditions: dict[str, Any] | None = None
    ) -> list[Any]:
        raise NoBindError()

    def insert(self, table: str, obj_data: dict[str, Any]) -> int | None:
        raise NoBindError()

    def migrate(self, table: str, columns: list[SQLColumn]) -> None:
        raise NoBindError()

    def update(self, table: str, obj_data: dict[str, Any], primary_key: str) -> None:
        raise NoBindError()

    def delete(self, table: str, key: str, value: Any) -> None:
        raise NoBindError()

    def execute_migration(self, migration: Migration, force: bool = False) -> None:
        raise NoBindError()



from __future__ import annotations

import pickle
from typing import TYPE_CHECKING, Any, TypeVar

from dbtogo.datatypes import SQLColumn

T = TypeVar("T", bound="DBModel")

if TYPE_CHECKING:
    from dbtogo.dbmodel import DBModel


class GeneralSQLSerializer:
    def _get_col_type(self, col: dict[str, str]) -> str:
        has_format = col["type"] == "string" and "format" in col.keys()
        return col["format"] if has_format else col["type"]

    def _get_column_schema(self, name: str, column: dict) -> SQLColumn:
        default = column.get("default", None)
        nullable = False

        if "anyOf" in column.keys():
            types = column["anyOf"]

            if len(types) > 2:
                raise TypeError("Field has to have a clear type")

            if types[0]["type"] != "null" and types[1]["type"] != "null":
                raise TypeError("Invalid field type")

            if types[0]["type"] == "null" and types[1]["type"] == "null":
                raise TypeError("Invalid field type")

            nullable = True
            column = types[0] if types[0]["type"] != "null" else types[1]

        str_type = self._get_col_type(column)

        return SQLColumn(name, str_type, nullable, default)

    def _standardise_schema_col(self, col: SQLColumn) -> SQLColumn:
        basics = ["integer", "string", "number", "boolean", "date-time"]
        if col.datatype in basics:
            return col

        col.datatype = "bytes"

        if col.default is not None:
            col.default = pickle.dumps(col.default)

        return col

    def serialize_schema(
        self,
        classname: str,
        schema: dict,
        primary: str | None = None,
        unique: list[str] = [],
    ) -> list[SQLColumn]:
        if "properties" in schema.keys():
            props = schema["properties"]
        else:
            props = schema["$defs"][classname]["properties"]

        cols = []
        for prop in props:
            raw_col = self._get_column_schema(prop, props[prop])
            standard_col = self._standardise_schema_col(raw_col)

            if standard_col.name == primary:
                standard_col.primary_key = True
                standard_col.nullable = False

            if standard_col.name in unique:
                standard_col.unique = True

            cols.append(standard_col)

        return cols

    def serialize_object(self, obj: DBModel, no_bind: bool = False) -> dict[str, Any]:
        columns = self.serialize_schema(obj.__class__.__name__, obj.model_json_schema())
        obj_data = {}

        for col in columns:
            if col.datatype != "bytes":
                obj_data[col.name] = obj.__dict__.get(col.name, None)
                continue

            raw_obj = obj.__dict__.get(col.name, None)
            obj_data[col.name] = pickle.dumps(raw_obj)

        return obj_data

    def deserialize_object(self, cls: type[T], obj_data: tuple[Any]) -> T:
        columns = self.serialize_schema(cls.__name__, cls.model_json_schema())
        values = {}

        for i, col in enumerate(columns):
            value = obj_data[i]
            if col.datatype == "bytes":
                value = pickle.loads(value)

            values[col.name] = value

        result = cls(**values)
        return result
